/****************************************************************************
 This file is a part of Purp (Pugo RPM) - a ncurses-based RPM-handler
 Copyright (C) 1998 - Anders Karlsson <pugo@lysator.liu.se>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.         
 
 For latest information and version of Purp, look at the Purp homepage:
 <http://www.lysator.liu.se/purp/>
 
 Please send comments and bug-reports to: <pugo@lysator.liu.se>
****************************************************************************/
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <curses.h>
#include <fcntl.h>

#include "main.h"
#include "linklist.h"
#include "dir.h"
#include "cmdwin.h"
#include "textwindow.h"
#include "table.h"
#include "tools.h"
#include "install.h"
#include "progresswin.h"

extern struct    tablewin table[];

extern int       panel;
extern char      *rootdir;
extern char      *availdir;

extern listelement *listpointer[2];
extern listelement *topListpointer[2];

extern cmdwin * cmdline;  /* The command-line-object */ 
extern rpmdb grpmdb;      /* for the rpm-database */

progresswin *progressbar;
int progressCount = 0;


static void UpdateInstallProgress(const unsigned long amount, const unsigned long total)
{
   if( (((float) amount) / total * 100) >= progressCount )
     {
	progressCount++;
	progressbar->Update( +1 );
     }

   if( amount == total )
     LockTTY( 1 );
}    

/**************************************************************
 * InstallPackages
 * 
 * Install marked packages
 **************************************************************/

short InstallPackages( short operation, short installTemporary )
{
   listelement *tmpPtr, *tmpPtr2, *tmpPtr3;
   struct textwindow *twin;
   char tmpStr[ 128 ], pkgName[ 256 ];
   int fd, dep_fd;
   char * netsharedPath = NULL;   
   short confirmYes = 0, stopInstall = 0, forceInstall = 0, op_panel, result;
   Header h;
   int offset;
   int isSource;      
   rpmNotifyFunction fn = UpdateInstallProgress;
   /* Dependencies-variables */
   rpmDependencies rpmdep;
   struct rpmDependencyConflict * conflicts;
   int numConflicts;  
   
   
   /* Choose which list to work on */
   if( operation == UNINSTALL )
     {
	tmpPtr = topListpointer[ INSTALLED ];
	op_panel = INSTALLED;
     }
   else
     {
	tmpPtr = topListpointer[ AVAILABLE ];
	op_panel = AVAILABLE;
     }
   
   
#ifdef DEBUG_TO_TERM
   switch( operation )
     {
      case INSTALL:
	TermWrite( "InstallPackages()" );
	break;
      case UPGRADE:
	TermWrite( "UpgradePackages()" );     
	break;
      case UNINSTALL:
	TermWrite( "ErasePackages()" );     
	break;
     }
#endif
     
   netsharedPath = rpmGetVar(RPMVAR_NETSHAREDPATH);
   
   
   /* Initiate dependencies-structure */
   rpmdep = rpmdepDependencies( grpmdb );  
   

   /* Create the confirm-window */
   textwin *listwin;
   switch( operation )
     {
      case INSTALL:
	listwin = new textwin( "Please confirm installation of following packages", "", "", 1, NOWRAP); 
	break;
      case UPGRADE:
	listwin = new textwin( "Please confirm upgrade of following packages", "", "", 1, NOWRAP); 
	break;
      case UNINSTALL:
	listwin = new textwin( "Please confirm uninstallation of following packages", "", "", 1, NOWRAP); 
	break;
     }
   
   /* Jump to element after the dummy */
   if( tmpPtr->straightlink != NULL ) 
     tmpPtr = (listelement *)tmpPtr->straightlink;
   
   /* Show all marked files in the confirm-window */
   while( tmpPtr != NULL )
     {
	if( tmpPtr->is_marked )
	  {
	     sprintf( pkgName, "%s%s", availdir, tmpPtr->filename );

	     if( operation <= UPGRADE )
	       {
		  /* open file to get header for dependencies-check */
		  if ((fd = open( pkgName, O_RDONLY)) >= 0) 
		    {
		       if( ( rpmReadPackageHeader( fd, &h, &isSource, NULL, NULL) ) == 0 )
			 {
			    /* Add dependencies information */
			    if( operation == INSTALL )
				 rpmdepAddPackage( rpmdep, h, &pkgName );
			    else
			         rpmdepUpgradePackage( rpmdep, h, &pkgName );
			 }
		       close(fd);   
		    }
	       }
	     else
	       {
		  /* Add removance of the package to dependencies info */
		  offset = FindPackageOffset( tmpPtr );
		  rpmdepRemovePackage( rpmdep, offset );
	       }

	     
	     if( installTemporary )
	       tmpPtr2 = (listelement *)tmpPtr;
	     
	     sprintf( tmpStr, "%s/%s-%s-%s\n", 
		     tmpPtr->group, 
		     tmpPtr->name, 
		     tmpPtr->version, 
		     tmpPtr->release );
	     
	     listwin->AddStr( tmpStr, NOWRAP );
	     
	  }
	tmpPtr = (listelement *)tmpPtr->straightlink;
     }
   
   
   /* Confirm installation! */
   cmdline->Draw( CONFIRMLINE );
   
   if( cmdline->HandleConfirmLine() == TRUE )
     confirmYes = 1;
   
   delete listwin;


   /* The user confirmed! Time to do some business! */
   if( confirmYes )
     {
	/* Summarize dependencies-check */
	if( rpmdepCheck( rpmdep, &conflicts, &numConflicts) )
	  {
#ifdef DEBUG_TO_TERM
	     TermWrite("<---- Dependencies conflict! ---->");
#endif

	     stopInstall = 1;
	  }
	
	if( ! stopInstall && conflicts )
	  {
#ifdef DEBUG_TO_TERM
	     TermWrite("<---- Dependencies conflict! ---->");
#endif
	     result = HandleDependenyConflicts( operation, tmpPtr, conflicts, numConflicts );
	     rpmdepFreeConflicts( conflicts, numConflicts );  
	     if( result == DEP_FORCE ) forceInstall = 1;
	     stopInstall = 1;
	  }

	/* Free the dependencies-structure */
	rpmdepDone( rpmdep );
	
	if( ! stopInstall || forceInstall )
	  {
	     /* Choose which list to work on 
	      * If we are going to install or upgrade then init the progressbar */
	     if( operation == UNINSTALL )
	       {
		  tmpPtr = topListpointer[ INSTALLED ];
	       }
	     else
	       {
		  tmpPtr = topListpointer[ AVAILABLE ];
		  progressbar = new progresswin( "", 100, 0 );
	       }
	     
	     /* Jump to the element after the dummy */
	     if( tmpPtr->straightlink != NULL ) 
	       tmpPtr = (listelement *)tmpPtr->straightlink;
	     
	     /*******************************************************/
	     /* Run through the list and install the selected items */
	     /*                                                     */
	     while( tmpPtr != NULL )
	       {
		  if( tmpPtr->is_marked )
		    {
		       sprintf( pkgName, "%s/%s", availdir, tmpPtr->filename );
		       
		       /* VERIFY FIRST (If installation or upgrade) */
		       if( operation <= UPGRADE )
			 {
			    /* Reset progressbar (only used for installation & upgrade) */
			    sprintf( tmpStr, "%s-%s-%s", tmpPtr->name, tmpPtr->version, tmpPtr->release );
			    progressbar->Reset( tmpStr, 100, 0 );
			    progressCount = 0;
			    
			    if ((fd = open( pkgName, O_RDONLY)) >= 0) 
			      {
				 LockTTY( 1 );
				 if( operation == INSTALL )
				   {
				      if( rpmInstallPackage( rootdir, grpmdb, fd, NULL,
							    RPMINSTALL_TEST,
							    NULL, netsharedPath )) 
					{
#ifdef DEBUG_TO_TERM
					   TermWrite("<------- FILE CONFLICT! --------->");
#endif
					   stopInstall = 2;
					}
				   }
				 else
				   {
				      if( rpmInstallPackage( rootdir, grpmdb, fd, NULL,
							    RPMINSTALL_TEST |
							    RPMINSTALL_UPGRADE | 
							    RPMINSTALL_UPGRADETOOLD,
							    NULL, netsharedPath )) 
					{
#ifdef DEBUG_TO_TERM
					   TermWrite("<------- FILE CONFLICT! --------->");
#endif
					   stopInstall = 2;
					}
				   }
				 LockTTY( 0 );
				 close(fd);   
			      }
			 }
		       //		  else
		       //		    {
		       //		       /* Uninstallation */
		       //		    }
		       
		       
		       
		       /* If there is a conflict, then ask the user if we should
			* force the install. If it doesn't work with force, then 
			* use more force!  :-) */
		       
		       if( stopInstall == 2 )
			 forceInstall = ConfirmForceInstall( tmpPtr );
		       
		       /*******************************************
			*   INSTALL/UPGRADE/ERASE THE PACKAGE!!   *
			*******************************************/
		       
		       if( !stopInstall || forceInstall )
			 {
			    
			    if( operation <= UPGRADE )
			      {
				 if ((fd = open( pkgName, O_RDONLY)) >= 0) 
				   {
				      if( operation == INSTALL )
					{
#ifdef DEBUG_TO_TERM
					   sprintf( tmpStr, "Installing %s", tmpPtr->filename );
					   TermWrite( tmpStr );
#endif
					   if( rpmInstallPackage( rootdir, grpmdb, fd, NULL,
								 RPMINSTALL_REPLACEPKG | 
								 RPMINSTALL_REPLACEFILES,
								 fn, netsharedPath ))
					     {
						fprintf(stderr, "Error: could not install package!\n");
						exit(1);   
					     }
//					   UpdatePackagesCounter( tmpPtr->group, +1, INSTALLED );
//					   UpdatePackagesCounter( tmpPtr->group, -1, AVAILABLE );
					}
				      else
					{
#ifdef DEBUG_TO_TERM
					   sprintf( tmpStr, "Upgrading %s", tmpPtr->filename );
					   TermWrite( tmpStr );
#endif
					   if( rpmInstallPackage( rootdir, grpmdb, fd, NULL,
								 RPMINSTALL_UPGRADE | 
								 RPMINSTALL_UPGRADETOOLD,
								 fn, netsharedPath )) 
					     {
						fprintf(stderr, "Error: could not upgrade package!\n");
						exit(1);   
					     }
					}
				      LockTTY( 0 );
				      close(fd);
				   }
				 tmpPtr->is_marked = FALSE;
			      }
			    else
			      {
#ifdef DEBUG_TO_TERM
				 sprintf( tmpStr, "Uninstalling %s", tmpPtr->filename );
				 TermWrite( tmpStr );
#endif
				 /* Uninstall package */
				 offset = FindPackageOffset( tmpPtr );
				 
				 if( offset >= 0 )
				   {
				      if( rpmRemovePackage( rootdir, grpmdb, offset, 0 ) )
					{
					   fprintf(stderr, "Error: could not remove package!\n");
					   exit(1);   
					}
				      
				      /* Check if the package exists in the available-panel,
				       * if it does we want to set the is_installed-flag
				       * of that package to FALSE, so it will show up in
				       * that panel, available for installation */
				      
				   }
			      }
			    
			    /* If we really installed the package, then remove it from
			     * the marked / marked-size - info */
			    if( ! installTemporary )
			      {
				 table[ op_panel ].selected --;
				 table[ op_panel ].selected_size -= tmpPtr->size ;
			      }
			    
			    /* Operate on the list-structures */
			    switch( operation )
			      {
			       case INSTALL:
				 tmpPtr = (listelement *)MoveItem( topListpointer[INSTALLED], 
								   AVAILABLE, tmpPtr );
				 
				 break;
			       
			       case UPGRADE:
				 if( (tmpPtr3 = CheckPackageExists( INSTALLED, 
								   tmpPtr->name, 
								   NULL, NULL,
								   CASE_SENSITIVE)))
				   {
				      if( tmpPtr3->is_available )
					MoveItem( topListpointer[AVAILABLE], 
						  INSTALLED, tmpPtr3 );
				      else
					{
					   UpdatePackagesCounter( tmpPtr3->group, 
								 -1, INSTALLED );
					   RemoveItem( tmpPtr3, panel );
					}
				   }
				 tmpPtr = (listelement *)MoveItem( topListpointer[INSTALLED], 
								   AVAILABLE, tmpPtr );
				 break;
			       
			       case UNINSTALL:
				 if( tmpPtr->is_available )
				   tmpPtr = (listelement *)MoveItem( topListpointer[AVAILABLE], 
								     INSTALLED, tmpPtr );
				 else
				   {
				      UpdatePackagesCounter( tmpPtr->group, 
							    -1, INSTALLED );
				      tmpPtr = (listelement *)RemoveItem( tmpPtr, panel );
				   }
				 break;
			      }
			 }
		       else 
			 {
			    /* If the user aborted because of file-conflicts
			     * go to next item in the list */
			    tmpPtr = (listelement *)tmpPtr->straightlink;
			    
			    /* Remove the temporary marking of the package */
			    if( installTemporary )
			      tmpPtr2->is_marked = FALSE;
			 }
		       stopInstall = 0;
		    }
		  else
		    tmpPtr = (listelement *)tmpPtr->straightlink;
	       }
	
	     if( operation <= UPGRADE )
	       delete progressbar;   
	     
	     /* Move cursors.. And so on.. Bla bla bla.. Tired.. */
	     table[ INSTALLED ].new_option = 0;
	     table[ INSTALLED ].highest_item_on_list = 0;
	     table[ AVAILABLE ].new_option = 0;
	     table[ AVAILABLE ].highest_item_on_list = 0;
	  }
	else if( installTemporary )
	  {
	     /* If something made us not installing the package, then unselect it */
	     tmpPtr2->is_marked = FALSE;
	  }
	
	
	display_menu( listpointer, 0, table[ INSTALLED ].new_option );
	display_menu( listpointer, 1, table[ AVAILABLE ].new_option );
	DrawInfoWin();
	
     }
   else if( installTemporary )
     {
	/* If something made us not installing the package, then unselect it */
	tmpPtr2->is_marked = FALSE;
     }
}

   

short ConfirmForceInstall( listelement *tmpPtr )
{			 
   textwin *forcewin;
    char tmpStr[ 128 ];
  
   
   forcewin = new textwin( "File conflicts", "", "", 1, NOWRAP); 
   
   /* Make some noice to wake up the user.. */
   beep();

   sprintf( tmpStr, "\nInstallation of package: %s-%s-%s\n", 
	   tmpPtr->name, tmpPtr->version, tmpPtr->release );
   
   forcewin->AddStr( tmpStr, NOWRAP );
   forcewin->AddStr( "conflicts with files from other installed packages.\n\n", NOWRAP );

   forcewin->AddStr( "\nCommands:\n", NOWRAP );
   forcewin->AddStr( "  Force: Force operation, even if it overwrites files\n", NOWRAP );
   forcewin->AddStr( "  Abort: Abort installation of this package\n", NOWRAP );
   
   /* Confirm installation! */
   cmdline->Draw( FORCELINE );
   
   if( cmdline->HandleForceLine() == TRUE )
     {
	delete forcewin;
	return 1;
     }
   else
     {
	delete forcewin;
	return 0;
     }
}



short HandleDependenyConflicts( short operation, listelement *tmpPtr,  
			        struct rpmDependencyConflict * conflicts,
			        int numConflicts)
{			 
   textwin *depwin;
   char tmpStr[ 128 ];
   int i;
   short result;
   
   depwin = new textwin( "Dependencies conflicts", "", "", 1, NOWRAP); 
   
   /* Make some noice to wake up the user.. */
   beep();
   
   depwin->AddStr( "\n", NOWRAP );

   for (i = 0; i < numConflicts; i++) 
     {
	sprintf( tmpStr, "%s", conflicts[i].needsName );
	
	if( conflicts[i].needsFlags )
	  {
	     if (conflicts[i].needsFlags)
	       sprintf( tmpStr, "%s ", tmpStr);
	     
	     if (conflicts[i].needsFlags & RPMSENSE_LESS)
	       sprintf( tmpStr, "%s<", tmpStr);
	     if (conflicts[i].needsFlags & RPMSENSE_GREATER)
	       sprintf( tmpStr, "%s>", tmpStr);
	     if (conflicts[i].needsFlags & RPMSENSE_EQUAL)
	       sprintf( tmpStr, "%s=", tmpStr);
	     if (conflicts[i].needsFlags & RPMSENSE_SERIAL)
	       sprintf( tmpStr, "%sS", tmpStr);
	     
	     if (conflicts[i].needsFlags)
	       sprintf( tmpStr, "%s %s", tmpStr, conflicts[i].needsVersion ); 
	  }
	
	  sprintf( tmpStr, "%s is needed by %s-%s-%s\n", tmpStr, 
		  conflicts[i].byName, conflicts[i].byVersion, conflicts[i].byRelease);
		
	depwin->AddStr( tmpStr, NOWRAP );
     }

   depwin->AddStr( "\n\nCommands:\n", NOWRAP );
   depwin->AddStr( "  Force: Force operation, even if it breaks dependencies\n", NOWRAP );

/********************************** NOT YET IMPLEMENTED **************************
   if( operation <= UPGRADE )
     depwin->AddStr( "  Solve: Install (if available) the packages to solve the conflicts\n", NOWRAP );
   else 
     depwin->AddStr( "  Solve: Do not uninstall packages that are needed\n", NOWRAP );
*/ 
   
   depwin->AddStr( "  Abort: Abort this operation\n", NOWRAP );
   
   
   /* Confirm installation! */
   cmdline->Draw( DEPENDENCIESLINE );

   result = cmdline->HandleDependenciesLine();

   delete depwin;

   return result;
}

