/****************************************************************************
 This file is a part of Purp (Pugo RPM) - a ncurses-based RPM-handler
 Copyright (C) 1998 - Anders Karlsson <pugo@lysator.liu.se>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.         
 
 For latest information and version of Purp, look at the Purp homepage:
 <http://www.lysator.liu.se/purp/>
 
 Please send comments and bug-reports to: <pugo@lysator.liu.se>
****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <curses.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>

extern "C" { 
   #include <rpm/rpmlib.h>
} 

#include "linklist.h"
#include "dir.h"
#include "main.h"
#include "mainloop.h"
#include "table.h"
#include "tools.h"

#include "cmdwin.h"
#include "pathwin.h"


int oldstdout;
int oldstderr;

/* Define the two panels.. */
struct tablewin table[2];

bool        quitloop=FALSE;
WINDOW      *infowin;
int         panel;

listelement *listpointer[2];
listelement *topListpointer[2];

char * rootdir = "/";
char * rcfile = NULL, * queryFormat = NULL, * prefix = NULL; 
char *arch = NULL; 
char *os = NULL; 
int  building = 0, showrc = 0;
char *availdir;
rpmdb grpmdb;  /* for the rpm-database */
cmdwin *cmdline;
pathwin *pathline;
int tty[2];
short rpmdb_write_access = 1;



#ifdef DEBUG_TO_TERM
int termdev = -1;
#endif

//#####################################################################

int main(int argc, char **argv)
{
   char tmpstr[255];
   int errval;
   availdir = (char *)xmalloc( 512 );

   
   /* Install signal-handlers */
   signal( SIGSEGV, sigSEGVhandler );
   signal( SIGHUP, sigHUPhandler );
   signal( SIGTERM, sigTERMhandler );


   /* Interpret the arguments on the commandline */
   if( argc > 1 )
     {
	if( ScanCommandLine( argc, argv ) )
	  return( 0 );
     }
   else
     {
	strcpy( availdir, "" );
     }
   
#ifdef DEBUG_TO_TERM
   TermWrite( "" );
   TermWrite( "===================================================" );
   TermWrite( PURP_VERSION_STRING );
   TermWrite( "===================================================" );
   TermWrite( "" );
#endif
   
         
   /* Read the RPM-configs */
   if( rpmReadConfigFiles(rcfile, arch, os, building) )
     {
	fprintf(stderr, "could not read config-files!\n");
	exit(1);
     }
   
   
   if (showrc) 
     {
	rpmShowRC(stdout);
	  {
	     fprintf(stderr, "could not show config-files!\n");
	     exit(0);
	  }
     }     

   
   /* Try to open the RPM-database with write-access */
   if( rpmdbOpen( rootdir, &grpmdb, O_RDWR | O_EXCL, 0644) ) 
     {
	fprintf(stderr, "User doesn't have write-permission for the RPM-database!\n");
	fprintf(stderr, "Starting Purp in it's limited mode (for normal users). \n");
	
	/* The user can't write, open with read-access only */
	if( rpmdbOpen( rootdir, &grpmdb, O_RDONLY, 0644) ) 
	  {
	     fprintf(stderr, "Error: could not open rpm-database!\n");
	     exit(1);   
	  }
	rpmdb_write_access = 0;
     }
   
    
   /* Do some main-initialization */
   SetupGlobals();
   InitieraCurs();

   /* Quit with error if the terminal is smaller than 80*24
    * otherwise ncurses will be unhappy and Purp will look bad */ 
   if( (LINES < 24) || (COLS < 80) )
     {
	/* Delete ncurses-windows and finish our play with ncurses.. */
	delwin( table[ INSTALLED ].win );
	delwin( table[ AVAILABLE ].win );
	delwin( infowin );
	endwin();

	fprintf(stderr, "Error: Purp needs at least a terminal with 80 columns and 24 lines!\n");
	fprintf(stderr, "       Current values:  Columns: %d, Lines: %d\n", COLS, LINES );
	exit(1);   
     }

    /* Hide the cursor, it won't make anyone happy sitting in a corner flashing.. */
   errval = curs_set( 0 );
   
   /* Create the command-line-object. Show about-line while starting up */
   cmdline = new cmdwin( ABOUTLINE );
   
   /* Create the path-win-object */
   pathline = new pathwin();

   /* Create dummys */
   listpointer[ INSTALLED ] = CreateDummy( listpointer[ INSTALLED ] );
   listpointer[ AVAILABLE ] = CreateDummy( listpointer[ AVAILABLE ] );
      
   topListpointer[ INSTALLED ] = listpointer[ INSTALLED ];
   topListpointer[ AVAILABLE ] = listpointer[ AVAILABLE ];

   listpointer[ INSTALLED ] = ReadInstalledRPM( listpointer[ INSTALLED ] );
   listpointer[ AVAILABLE ] = ReadAvailableRPM( listpointer[ AVAILABLE ] );

   /* After the startup: Show the real command-line */
   cmdline->Draw( MAINLINE );
     
   /******************************************************************/
   /* The MAIN-LOOP that handles the major part of Purp's functions! */
   /******************************************************************/
   MainLoop( listpointer );
      
   /*************************************************/
   /* Close everything, free everything, halleluja! */
   /*************************************************/
  
   quitPurp( NULL );
}


/*************************************************/
/* Signal handlers                               */
/*************************************************/
void sigSEGVhandler( int sig )
{
#ifdef DEBUG_TO_TERM
     TermWrite( SIGSEGV_MESSAGE );
#endif
   quitPurp( SIGSEGV_MESSAGE );
}

void sigHUPhandler( int sig )
{
#ifdef DEBUG_TO_TERM
     TermWrite( SIGHUP_MESSAGE );
#endif
   quitPurp( SIGHUP_MESSAGE );
}

void sigTERMhandler( int sig )
{
#ifdef DEBUG_TO_TERM
     TermWrite( SIGTERM_MESSAGE );
#endif
   quitPurp( SIGTERM_MESSAGE );
}


/*************************************************/
/* Close everything, free everything, halleluja! */
/*************************************************/
void quitPurp( char * message )
{
   /* Close the RPM-database */
   rpmdbClose( grpmdb );

   /* destruct the command-line-object */
   delete cmdline;

   /* destruct the pathline */
   delete pathline;
			    
   /* Make the cursor visible */
   curs_set( 1 );

   /* Delete ncurses-windows and finish our play with ncurses.. */
   delwin( table[ INSTALLED ].win );
   delwin( table[ AVAILABLE ].win );
   delwin( infowin );
   endwin();
   
   /* Destroy the linked lists containing package-info, and free their memory */
   EraseList( topListpointer[ INSTALLED ] );
   EraseList( topListpointer[ AVAILABLE ] );

   free( availdir );
   
   #ifdef DEBUG_TO_TERM
     TermWrite( "Quiting Purp.. Good bye dear user!" );
     close( termdev );
   #endif
     
   /* Print error-message if there are any */
   if( message )
       fprintf( stderr, "\nPurp: %s\n", message );
	       
   /* Not much left to do.. Hope the user enjoyed Purp.. */
   exit(0);
}


/*************************************************
 * Setup global variables, prepare the lists    
 *************************************************/
void SetupGlobals()
{
   panel = 0;
   
   /* Initiate lists for the package-hierarchies */
   listpointer[ INSTALLED ] = NULL;
   listpointer[ AVAILABLE ] = NULL;

   table[ INSTALLED ].selected = 0;
   table[ AVAILABLE ].selected = 0;

   table[ INSTALLED ].selected_size = 0;
   table[ AVAILABLE ].selected_size = 0;
   
   table[ INSTALLED ].highest_item_on_list = 0;
   table[ AVAILABLE ].highest_item_on_list = 0;
}



/*************************************************
 * Scan and parse commandline
 *************************************************/

short ScanCommandLine( int argc, char ** argv )
{
   int count;
   short argConsumed = 0;
   short quitAfter = 0;
   
   if( argc > 1 )
     {
	for( count=1; count < argc; count++ )
	  {
	     if( argv[ count ][ 0 ] == '-' )
	       {
		  InterpretFlag( argv, &count, &quitAfter );
	       }
	     else 
	       {
		  if( count < argc )
		    {
		       strcpy( availdir, argv[ count ] ); 
		       break;
		    }
		  else
		    strcpy( availdir, "" );
	       }
	  }
     }
   
   return( quitAfter );
}


void InterpretFlag( char ** argv, int * count, short * quitAfter )
{
   char tmpStr[256];

   if( ! strcmp( argv[ *count ], "--help" ) )
     {
	PrintUsageInfo();
	*quitAfter = 1;
     }
   else if( ! strcmp( argv[ *count ], "--version" ) )
     {
	sprintf( tmpStr, "%s\n", PURP_VERSION_STRING );
	fprintf( stdout, tmpStr );
	*quitAfter = 1;
     }
   
   else if( ! strcmp( argv[ *count ], "--debug" ) )
     {
#ifdef DEBUG_TO_TERM
	termdev = open(  argv[ *count + 1 ], O_WRONLY );
#else
	fprintf( stdout, "This version of Purp isn't built with Debug-support.\n" );
	*quitAfter = 1;
#endif
	*count = *count + 1;
     }
   
   else
     {
	fprintf( stdout, "Unknown parameter: %s\n\n", argv[ *count ] );
	PrintUsageInfo();
	*quitAfter = 1;
     }
}


void PrintUsageInfo( void )
{
   fprintf( stdout, "Usage: purp [--option]... [package-dir]\n\n" );
   fprintf( stdout, "where [package-dir] is a directory containing rpm-packages\n\n" );
   fprintf( stdout, "options are:\n" );
   fprintf( stdout, "--help              display this help and exit\n" );
   fprintf( stdout, "--version           display version information and exit\n" );
#ifdef DEBUG_TO_TERM
   fprintf( stdout, "--debug <device>    device to which debuginfo is sent\n\n" );
#endif
}
